/*
 * This source code is a product of Sun Microsystems, Inc. and is provided
 * for unrestricted use.  Users may copy or modify this source code without
 * charge.
 *
 * SUN SOURCE CODE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING
 * THE WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 * Sun source code is provided with no support and without any obligation on
 * the part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 *
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS SOFTWARE
 * OR ANY PART THEREOF.
 *
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 *
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

/*
 * g723_40.c
 *
 * Description:
 *
 * g723_40_encoder (), g723_40_decoder ()
 *
 * These routines comprise an implementation of the CCITT G.723 40Kbps
 * ADPCM coding algorithm.  Essentially, this implementation is identical to
 * the bit level description except for a few deviations which
 * take advantage of workstation attributes, such as hardware 2's
 * complement arithmetic.
 *
 * The deviation from the bit level specification (lookup tables),
 * preserves the bit level performance specifications.
 *
 * As outlined in the G.723 Recommendation, the algorithm is broken
 * down into modules.  Each section of code below is preceded by
 * the name of the module which it is implementing.
 *
 */

#include "g72x.h"
#include "g72x_priv.h"

/*
 * Maps G.723_40 code word to ructeconstructed scale factor normalized log
 * magnitude values.
 */
static short	_dqlntab [32] = { -2048, -66, 28, 104, 169, 224, 274, 318,
				358, 395, 429, 459, 488, 514, 539, 566,
				566, 539, 514, 488, 459, 429, 395, 358,
				318, 274, 224, 169, 104, 28, -66, -2048 } ;

/* Maps G.723_40 code word to log of scale factor multiplier. */
static short	_witab [32] = { 448, 448, 768, 1248, 1280, 1312, 1856, 3200,
			4512, 5728, 7008, 8960, 11456, 14080, 16928, 22272,
			22272, 16928, 14080, 11456, 8960, 7008, 5728, 4512,
			3200, 1856, 1312, 1280, 1248, 768, 448, 448 } ;

/*
 * Maps G.723_40 code words to a set of values whose long and short
 * term averages are computed and then compared to give an indication
 * how stationary (steady state) the signal is.
 */
static short	_fitab [32] = { 0, 0, 0, 0, 0, 0x200, 0x200, 0x200,
			0x200, 0x200, 0x400, 0x600, 0x800, 0xA00, 0xC00, 0xC00,
			0xC00, 0xC00, 0xA00, 0x800, 0x600, 0x400, 0x200, 0x200,
			0x200, 0x200, 0x200, 0, 0, 0, 0, 0 } ;

static short qtab_723_40 [15] = { -122, -16, 68, 139, 198, 250, 298, 339,
				378, 413, 445, 475, 502, 528, 553 } ;

/*
 * g723_40_encoder ()
 *
 * Encodes a 16-bit linear PCM, A-law or u-law input sample and retuens
 * the resulting 5-bit CCITT G.723 40Kbps code.
 * Returns -1 if the input coding value is invalid.
 */
int	g723_40_encoder (int sl, G72x_STATE *state_ptr)
{
	short		sei, sezi, se, sez ;	/* ACCUM */
	short		d ;			/* SUBTA */
	short		y ;			/* MIX */
	short		sr ;			/* ADDB */
	short		dqsez ;			/* ADDC */
	short		dq, i ;

	/* linearize input sample to 14-bit PCM */
	sl >>= 2 ;		/* sl of 14-bit dynamic range */

	sezi = predictor_zero (state_ptr) ;
	sez = sezi >> 1 ;
	sei = sezi + predictor_pole (state_ptr) ;
	se = sei >> 1 ;			/* se = estimated signal */

	d = sl - se ;			/* d = estimation difference */

	/* quantize prediction difference */
	y = step_size (state_ptr) ;	/* adaptive quantizer step size */
	i = quantize (d, y, qtab_723_40, 15) ;	/* i = ADPCM code */

	dq = reconstruct (i & 0x10, _dqlntab [i], y) ;	/* quantized diff */

	sr = (dq < 0) ? se - (dq & 0x7FFF) : se + dq ; /* reconstructed signal */

	dqsez = sr + sez - se ;		/* dqsez = pole prediction diff. */

	update (5, y, _witab [i], _fitab [i], dq, sr, dqsez, state_ptr) ;

	return i ;
}

/*
 * g723_40_decoder ()
 *
 * Decodes a 5-bit CCITT G.723 40Kbps code and returns
 * the resulting 16-bit linear PCM, A-law or u-law sample value.
 * -1 is returned if the output coding is unknown.
 */
int	g723_40_decoder	(int i, G72x_STATE *state_ptr)
{
	short		sezi, sei, sez, se ;	/* ACCUM */
	short		y ;			/* MIX */
	short		sr ;			/* ADDB */
	short		dq ;
	short		dqsez ;

	i &= 0x1f ;			/* mask to get proper bits */
	sezi = predictor_zero (state_ptr) ;
	sez = sezi >> 1 ;
	sei = sezi + predictor_pole (state_ptr) ;
	se = sei >> 1 ;			/* se = estimated signal */

	y = step_size (state_ptr) ;	/* adaptive quantizer step size */
	dq = reconstruct (i & 0x10, _dqlntab [i], y) ;	/* estimation diff. */

	sr = (dq < 0) ? (se - (dq & 0x7FFF)) : (se + dq) ; /* reconst. signal */

	dqsez = sr - se + sez ;		/* pole prediction diff. */

	update (5, y, _witab [i], _fitab [i], dq, sr, dqsez, state_ptr) ;

	return arith_shift_left (sr, 2) ;	/* sr was of 14-bit dynamic range */
}

